use std::str::FromStr;

use baseline::Baseline;
use palette::FromColor;
use scheme_setting::{ColorExpand, WACGSetting};
use serde::Serialize;
use strum::IntoEnumIterator;
use wasm_bindgen::{prelude::wasm_bindgen, JsValue};

use crate::{errors, parse_option_to_oklch, parse_to_oklch};

use super::SchemeExport;

mod baseline;
mod color_set;
mod neutral_swatch;
mod scheme_setting;

pub use scheme_setting::{ColorShifting, SchemeSetting};

#[derive(Debug, Clone, Serialize)]
pub struct QScheme {
    pub light: Baseline,
    pub dark: Baseline,
}

impl QScheme {
    pub fn new(
        primary: &str,
        danger: &str,
        success: &str,
        warning: &str,
        info: &str,
        foreground: &str,
        background: &str,
        setting: SchemeSetting,
    ) -> Result<Self, errors::ColorError> {
        let primary = parse_to_oklch!(primary);
        let danger = parse_to_oklch!(danger);
        let success = parse_to_oklch!(success);
        let warning = parse_to_oklch!(warning);
        let info = parse_to_oklch!(info);
        let foreground = parse_to_oklch!(foreground);
        let background = parse_to_oklch!(background);
        Ok(Self {
            light: Baseline::new(
                &primary,
                &danger,
                &success,
                &warning,
                &info,
                &foreground,
                &background,
                setting.clone(),
            ),
            dark: Baseline::new(
                &(primary * setting.dark_convert),
                &(danger * setting.dark_convert),
                &(success * setting.dark_convert),
                &(warning * setting.dark_convert),
                &(info * setting.dark_convert),
                &(foreground * setting.dark_convert),
                &(background * setting.dark_convert),
                setting.clone(),
            ),
        })
    }

    pub fn custom(
        primary: &str,
        secondary: Option<&str>,
        tertiary: Option<&str>,
        accent: Option<&str>,
        danger: &str,
        success: &str,
        warning: &str,
        info: &str,
        foreground: &str,
        background: &str,
        setting: SchemeSetting,
    ) -> Result<Self, errors::ColorError> {
        let primary = parse_to_oklch!(primary);
        let secondary = parse_option_to_oklch!(secondary);
        let tertiary = parse_option_to_oklch!(tertiary);
        let accent = parse_option_to_oklch!(accent);
        let danger = parse_to_oklch!(danger);
        let success = parse_to_oklch!(success);
        let warning = parse_to_oklch!(warning);
        let info = parse_to_oklch!(info);
        let foreground = parse_to_oklch!(foreground);
        let background = parse_to_oklch!(background);

        Ok(Self {
            light: Baseline::custom(
                &primary,
                &secondary,
                &tertiary,
                &accent,
                &danger,
                &success,
                &warning,
                &info,
                &foreground,
                &background,
                setting.clone(),
            ),
            dark: Baseline::custom(
                &(primary * setting.dark_convert),
                &secondary.map(|color| color * setting.dark_convert),
                &tertiary.map(|color| color * setting.dark_convert),
                &accent.map(|color| color * setting.dark_convert),
                &(danger * setting.dark_convert),
                &(success * setting.dark_convert),
                &(warning * setting.dark_convert),
                &(info * setting.dark_convert),
                &(foreground * setting.dark_convert),
                &(background * setting.dark_convert),
                setting.clone(),
            ),
        })
    }
}

impl SchemeExport for QScheme {
    fn output_css_variables(&self) -> String {
        let mut variables = Vec::new();

        variables.extend(self.light.to_css_variables("light"));
        variables.extend(self.dark.to_css_variables("dark"));

        variables.join("\n")
    }

    fn output_scss_variables(&self) -> String {
        let mut variables = Vec::new();

        variables.extend(self.light.to_scss_variables("light"));
        variables.extend(self.dark.to_scss_variables("dark"));

        variables.join("\n")
    }

    fn output_javascript_object(&self) -> String {
        let mut javascript_object = Vec::new();

        javascript_object.push("{".to_string());
        javascript_object.push("  light: {".to_string());
        javascript_object.extend(
            self.light
                .to_javascript_fields()
                .into_iter()
                .map(|c| format!("    {}", c))
                .collect::<Vec<_>>(),
        );
        javascript_object.push("  },".to_string());
        javascript_object.push("  dark: {".to_string());
        javascript_object.extend(
            self.dark
                .to_javascript_fields()
                .into_iter()
                .map(|c| format!("    {}", c))
                .collect::<Vec<_>>(),
        );
        javascript_object.push("}".to_string());

        javascript_object.join("\n")
    }
}

#[wasm_bindgen]
pub fn q_scheme_color_expanding_methods() -> Result<JsValue, String> {
    let methods = ColorExpand::iter()
        .map(|variant| {
            serde_json::json!({
                "label": variant.label(),
                "value": variant.to_string(),
            })
        })
        .collect::<Vec<_>>();

    serde_wasm_bindgen::to_value(&methods).map_err(|e| e.to_string())
}

#[wasm_bindgen]
pub fn q_scheme_wacg_settings() -> Result<JsValue, String> {
    let settings = WACGSetting::iter()
        .map(|setting| {
            serde_json::json!({
                "label": setting.label(),
                "value": setting.to_string(),
            })
        })
        .collect::<Vec<_>>();

    serde_wasm_bindgen::to_value(&settings).map_err(|e| e.to_string())
}
